from enum import Enum
from re import Pattern
import re

class TokenClass(Enum):
    NUMBER = 'Number'
    IDENTIFIER = 'Identifier'
    KEY_VALUE_PAIR = 'KeyValuePair'
    DATE_TIME = 'DateTime'
    # Location, Path, Socket, IP, URL
    LOCATION = 'Location'
    VARIABLE = 'Variable'
    SYMBOL = 'Symbol'
    PUNCTUATION = 'Punctuation'
    MISC = 'Misc'
    UNKNOWN = 'Unknown'


class _RegexPatternEntry():
    def __init__(self, pattern: str, token_class: TokenClass, description: str):
        self.pattern: Pattern = re.compile(pattern)
        self.token_class: TokenClass = token_class
        self.description: str = description

class RegexTokenClassMatcher:

    def __init__(self):
        self.regex_list: list[_RegexPatternEntry] = self._fill_regex_list()
 
 
    def token_classes(self, tokens: list[str]) -> list[TokenClass]:
        return [self.token_class(token) for token in tokens]

    def token_class(self, token: str) -> TokenClass:
        token_class=TokenClass.UNKNOWN
        for rmp in self.regex_list:
            if rmp.pattern.fullmatch(token):
                token_class=rmp.token_class #(rmp.token_class, rmp.description)
                break
        return token_class


    ######################
    # Regex List
    ######################

    def _fill_regex_list(self) -> list[_RegexPatternEntry]:
        regex_list: list[_RegexPatternEntry] = []

        # try casting number into float should be done in KnownWordsDetector already
        # Numbers
        # e.g. '123', '123.45', '-123.45', '1e.0', '1e-0'
        # This is done by test for float instead
        # regex_list.append(
        #     RegexMaskPattern(
        #         r'-?\d+(\.\d+)?([eE]-?\d+)?',
        #         RegexMaskPattern.T_mask.NUMBER,
        #         'Numbers'
        #     ))

        # key value pairs
        regex_list.append(_RegexPatternEntry(
            r'\w+=\w+',
            TokenClass.KEY_VALUE_PAIR,
            'key-value-pairs simple'
        ))
        regex_list.append(_RegexPatternEntry(
            r'[^=]+=[^=]+',
            TokenClass.KEY_VALUE_PAIR,
            'key-value-pairs greedy'
        ))

        # TODO Note, these two will even match e.g. "=x" and "x="
        # unclean key-value pairs
        # if contains only one '='
        regex_list.append(_RegexPatternEntry(
            r'^[^=]*=[^=]+$',
            TokenClass.KEY_VALUE_PAIR,
            'key-value-pairs incomplete left'
        ))
        regex_list.append(_RegexPatternEntry(
            r'^[^=]+=[^=]*$',
            TokenClass.KEY_VALUE_PAIR,
            'key-value-pairs incomplete right'
        ))

        # multiple key-value pairs
        # comma separated

        regex_list.append(
            _RegexPatternEntry(
                r'[^=]+=[^=]+(,[^=]+=[^=]+)+',
                TokenClass.KEY_VALUE_PAIR,
                'key-value-pairs comma-separated'
            ))
        # Version numbers
        # v0.100', 'v0.97', 'v6.1.7601.23505',
        regex_list.append(
            _RegexPatternEntry(
                r'v\.?\d+(\.\d+[a-z]?)+',
                TokenClass.IDENTIFIER,
                'version numbers'
            ))

        # UUID
        # e.g. '12345678-1234-5678-1234-567812345678'
        regex_list.append(
            _RegexPatternEntry(
                r'[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}',
                TokenClass.IDENTIFIER,
                'UUID'
            ))

        # Hexadecimal numbers
        regex_list.append(
            _RegexPatternEntry(
                r'0x[0-9a-fA-F]+',
                TokenClass.NUMBER,
                'hexadecimal numbers prefixed'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'[0-9a-fA-F]{8}',
                TokenClass.NUMBER,
                'hexadecimal numbers without prefix, 8 digits'
            ))

        # Sockets
        regex_list.append(
            _RegexPatternEntry(
                r'(\d{1,3}\.){3}\d{1,3}:\d+',
                TokenClass.LOCATION,
                'sockets (ip+port)'
            ))

        # IP addresses
        regex_list.append(
            _RegexPatternEntry(
                r'(\d{1,3}\.){3}\d{1,3}',
                TokenClass.LOCATION,
                'IP addresses'
            ))

        # MAC-Addresses
        regex_list.append(
            _RegexPatternEntry(
                r'([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}',
                TokenClass.IDENTIFIER,
                'MAC addresses'
            ))

        # Filenames
        import mimetypes
        # len(mimetypes.types_map) #1032
        common_file_extensions = '|'.join(
            [k[1:] for k in mimetypes.types_map.keys()])

        regex_list.append(
            _RegexPatternEntry(
                r'[^\./\\]+\.('+common_file_extensions+')',
                TokenClass.LOCATION,
                'file names'
            ))

        # TODO Here are some sockets that begins with slash ;/

        # Pathes

        regex_list.append(
            _RegexPatternEntry(
                r'(/[^/ ]*)+/?',
                TokenClass.LOCATION,
                'Linux paths'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'([a-zA-Z]:)?(\\[^\\ ]*)+\\{,2}',
                TokenClass.LOCATION,
                'Windows paths'
            ))

        # TODO: This matches packages, pathes and filenames also

        # regex from: https://stackoverflow.com/questions/3809401/what-is-a-good-regular-expression-to-match-a-url
        # url_regex=re.compile(r'[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)')
        # url_regex=re.compile(r'([a-zA-Z]+://)?[a-zA-Z0-9\.-]+\.[a-zA-Z]{2,6}(:[0-9]{1,5})?(/.*)?')
        regex_list.append(
            _RegexPatternEntry(
                # r'([a-zA-Z]+://)?[a-zA-Z0-9][a-zA-Z0-9\.-]*\.[a-zA-Z]+[a-zA-Z0-9]+(:[0-9]{1,5})?(/.*)?',
                # Assumption: Most likely url contain only lowercase letters
                r'([a-zA-Z]+://)?[a-z0-9][a-zA-Z0-9\.-]*\.[a-zA-Z]+[a-zA-Z0-9]+(:[0-9]{1,5})?(/.*)?',
                TokenClass.LOCATION,
                'URLs'
            ))
        # Leftover packages

        # url with lower AND uppercase? AND underscore

        # OR? Combine urls and packages as LocationUrlOrPackage ?

        # TODO

        # Times
        # 2016/9/27:20:40:54.787
        regex_list.append(
            _RegexPatternEntry(
                r'\d{4}/\d{1,2}/\d{1,2}:\d{1,2}:\d{2}:\d{2}.\d+',
                TokenClass.DATE_TIME,
                'date-times slash'
            ))

        # 2017-07-03_13,48,39.308188
        regex_list.append(
            _RegexPatternEntry(
                r'\d{4}-\d{1,2}-\d{1,2}_\d{1,2},\d{2},\d{2}.\d+',
                TokenClass.DATE_TIME,
                'date-times dash milliseconds'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'\d{4}-\d{1,2}-\d{1,2}_\d{1,2},\d{2},\d{2}',
                TokenClass.DATE_TIME,
                'date-times dash seconds'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'\d{4}-\d{1,2}-\d{1,2}',
                TokenClass.DATE_TIME,
                'dates dash'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'\d{4}/\d{1,2}/\d{1,2}',
                TokenClass.DATE_TIME,
                'dates slash'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'\d{1,2}:\d{2}:\d{2}.\d+',
                TokenClass.DATE_TIME,
                'times float seconds'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'\d{1,2}:\d{2}:\d{2}',
                TokenClass.DATE_TIME,
                'times seconds'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'\d{1,2}:\d{2}',
                TokenClass.DATE_TIME,
                'times minutes'
            ))

        # TODO: Find unix timestamps as integer of a specific interval BEFORE numbers are matched

        # Now it's getting blurred

        # Look like normal words "Hello"
        regex_list.append(
            _RegexPatternEntry(
                r'[A-Z][a-z]+',
                TokenClass.VARIABLE,
                'words leading uppercase'
            ))

        # only uppercase words
        regex_list.append(
            _RegexPatternEntry(
                r'[A-Z]+',
                TokenClass.VARIABLE,
                'uppercase words'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'[a-z]+',
                TokenClass.VARIABLE,
                'lowercase words'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'[a-zA-Z]+\d+',
                TokenClass.VARIABLE,
                'words with trailing digits'
            ))

        # camelCase words

        regex_list.append(
            _RegexPatternEntry(
                r'[a-zA-Z]+([A-Z][a-z]+)+',
                TokenClass.VARIABLE,
                'camelCase words'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'[a-zA-Z]+([A-Z][a-z]+)+\d+',
                TokenClass.VARIABLE,
                'camelCase words with trailing digits'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'_+[a-zA-Z]+([A-Z][a-z]+)+',
                TokenClass.VARIABLE,
                'camelCase words with leading underscore'
            ))

        # snake_case (underscore separated) words

        regex_list.append(
            _RegexPatternEntry(
                r'[a-z]+(_[a-z]+)+',
                TokenClass.VARIABLE,
                'snake_case words'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'[a-z]+(_[a-z]+)+\d+',
                TokenClass.VARIABLE,
                'snake_case words with trailing digits'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'_+[a-z]+(_[a-z]+)+',
                TokenClass.VARIABLE,
                'snake_case words with leading underscore'
            ))

        # kebap (dash separated) words

        regex_list.append(
            _RegexPatternEntry(
                r'[a-z]+(-[a-z]+)+',
                TokenClass.VARIABLE,
                'kebap words only lowercase'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'[a-zA-Z]+(-[a-zA-Z]+)+',
                TokenClass.VARIABLE,
                'kebap words (lower and uppercase)'
            ))

        # words with underscore, digits and camelCase
        # job_1445144423722_0020Job
        regex_list.append(
            _RegexPatternEntry(
                r'[a-z]+[a-zA-Z0-9_-]+',
                TokenClass.VARIABLE,
                'mixed leading lowercase'
            ))

        regex_list.append(
            _RegexPatternEntry(
                r'[A-Z]+[a-zA-Z0-9_-]+',
                TokenClass.VARIABLE,
                'mixed leading uppercase'
            ))

        # more key-value with colon BUT word characters on the left

        regex_list.append(
            _RegexPatternEntry(
                r'(\w+):[^:]+',
                TokenClass.KEY_VALUE_PAIR,
                'key-value-pairs with colon'
            ))
        # unit values
        # 2.5MB, 2.5MB/s, 1024Kbytes, 3.60GHz, 300ms

        regex_list.append(
            _RegexPatternEntry(
                r'\d+(\.\d+)?[a-zA-Z/]+',
                TokenClass.NUMBER,
                'unit values'
            ))
        # Version number without 'v' prefix, but at least two dots, since numbers are already covered
        # Note: After numbers AND ip addresses AND units

        regex_list.append(
            _RegexPatternEntry(
                r'\d+(\.\d+[a-z]?)+',
                TokenClass.IDENTIFIER,
                'version numbers without v'
            ))
        # collect leftover hex-values
        # Numbers and words are already covered

        regex_list.append(
            _RegexPatternEntry(
                r'[0-9a-fA-F]+',
                TokenClass.NUMBER,
                'hexadecimal numbers leftover'
            ))
        
        # hex-values with seperators 
        # 00:06:6a:00:01:00:02:15 
        regex_list.append(
            _RegexPatternEntry(
                r'([0-9a-fA-F]{2}:)+[0-9a-fA-F]{2}',
                TokenClass.NUMBER,
                'hexadecimal numbers leftover with colons'
            ))


        # punctuation
        #import string
        #string.punctuation
        #!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~
        # These punctuation mess up SYM and PUNCT, so they were separated
        
        punct_chars=re.escape(r""".,;:!?()[]{}_…“”‘’"'/\|·«»`~¿¡•""")
        
        regex_list.append(
            _RegexPatternEntry(
                r'['+punct_chars+']',
                TokenClass.PUNCTUATION,
                'single punctuation'
            ))

        # multiple punctuation
        regex_list.append(
            _RegexPatternEntry(
                r'['+punct_chars+']+',
                TokenClass.PUNCTUATION,
                'multiple punctuation'
            ))
        
        # symbols
        sym_chars=re.escape(r"""+-=*^%$&§¤#@<>©®™°±×÷√∞∑∏∫∆µπΩ≠≈∈∩∪⊂⊃∅∇⊕⊗⇒⇔""")
        regex_list.append(
            _RegexPatternEntry(
                r'['+sym_chars+']',
                TokenClass.SYMBOL,
                'single symbol'
            ))
        # multiple symbols
        regex_list.append(
            _RegexPatternEntry(
                r'['+sym_chars+']+',
                TokenClass.SYMBOL,
                'multiple symbols'
            ))

        return regex_list




from abc import ABC, abstractmethod

import logging
logger = logging.getLogger(__name__)

class AbstractPosTagger(ABC):
    @abstractmethod
    def pos_tag(self,tokens:list[str], tagset='upos')->list[str]:
        pass

    def pos_tag_upos(self,tokens:list[str])->list[str]:
        return self.pos_tag(tokens, tagset='upos')

    def pos_tag_ptb(self,tokens:list[str])->list[str]:
        return self.pos_tag(tokens, tagset='ptb')


from setuptools import setup
from setuptools.command.install import install
import nltk

class PostInstallCommand(install):
    def run(self):
        install.run(self)
        # NLTK dependencies
        corpora_deps = ['words', 'stopwords', 'wordnet']
        for dependency in corpora_deps:
            try:
                nltk.data.find(f'corpora/{dependency}')
            except LookupError:
                print(f"Did not found corpora '{dependency}'. Downloading...")
                nltk.download(dependency)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='poslog',
    version='0.6',
    author='Kilian Dangendorf',
    description='PosLog: A CRF-based Part-of-Speech Tagger for Log Messages',
    long_description=long_description,
    long_description_content_type='text/markdown',
    #url='https://github.com/kiliandangendorf/poslog',
    project_urls={
        'GitHub': 'https://github.com/kiliandangendorf/poslog',
    },
    packages=['poslog', 'poslog.words'],
    include_package_data=True,
    package_data={
        'poslog': ['models/pos_log_upos_crf_10k_model.pkl', 'words/*.txt'],
    },
    install_requires=['nltk', 'sklearn-crfsuite'],
    cmdclass={
        'install': PostInstallCommand,
    },
    python_requires=">=3.11",

)


from enum import Enum
from nltk.corpus import stopwords
from nltk.corpus import wordnet
from nltk.corpus import words
import nltk

from pathlib import Path

import logging
logger = logging.getLogger(__name__)

class WordKind(Enum):
    STOP_WORD = 'stopword'
    WORD_NET = 'wordnet'
    WORDS_DICTIONARY = 'words_dictionary'
    DOMAIN_WORD = 'domain_word'
    NUMBER = 'number'
    UNKNOWN = 'unknown'

class KnownWordsDetector:

    def __init__(self, case_sensitive: bool = False):
        corpora_deps = ['words', 'stopwords', 'wordnet']
        for dependency in corpora_deps:
            try:
                nltk.data.find(f'corpora/{dependency}')
            except LookupError:
                logger.info(f"Initializing {self.__class__.__name__}")
                logger.info(f"Did not found corpora '{dependency}'. Downloading...")
                nltk.download(dependency)

        self.case_sensitive = case_sensitive
        # dict lookup is faster than list lookup
        self.words_dictionary: dict[str, None] = dict.fromkeys(
            [w.lower() for w in words.words()], None)
        self.stopwords_dict: dict[str, None] = dict.fromkeys(
            stopwords.words('english'), None)
        self.domain_words: dict[str, None] = self._read_domain_words()

    ######################
    # Known Words
    ######################

    def kind_of_known_words(self, tokens: list[str]) -> list[WordKind]:
        return [self.kind_of_known_word(token) for token in tokens]
    
    def kind_of_known_word(self, token: str) -> WordKind:
        if token.lower() in self.stopwords_dict:
            return WordKind.STOP_WORD
        # since this list build from words not in wordnet, we can do it before wordnet
        if self.is_domain_word(token):
            return WordKind.DOMAIN_WORD
        if self._is_number(token):
            return WordKind.NUMBER
        if wordnet.synsets(token):
            return WordKind.WORD_NET
        if token.lower() in self.words_dictionary:
            return WordKind.WORDS_DICTIONARY
        return WordKind.UNKNOWN

    # Note: Some integer numbers are already covered by wordnet
    def is_known_word(self, token: str):
        if self.kind_of_known_word(token) != WordKind.UNKNOWN:
            return True
        return False

    def _is_number(self, token: str) -> bool:
        try:
            # Remove common thousands separators (', and ')
            clean_t = token.replace(",", "").replace("'", "").replace(" ", "")
            float(clean_t)
            return True
        except:
            return False

    ######################
    # Domain Words
    ######################

    def _read_domain_words(self):
        DOMAIN_WORDS_FILENAME = "known_domain_words.txt"
        domain_words_file = Path(__file__).parent / DOMAIN_WORDS_FILENAME
        if not domain_words_file.exists():
            raise FileNotFoundError(
                f"Domain words file {domain_words_file} does not exist.")
        # load line wise from file
        with domain_words_file.open('r') as file:
            if self.case_sensitive:
                return [line.strip() for line in file.readlines()]
            else:
                return [line.strip().lower() for line in file.readlines()]

    def is_domain_word(self, word: str) -> bool:
        if not self.case_sensitive:
            word = word.lower()
        return word in self.domain_words
    



from nltk import word_tokenize
import re

######################
# Tokenizer
######################

class PosLogTokenizer:
    # Contractions from https://stackoverflow.com/a/19794953
    # Always choose the last ambiguous option
    CONTRACTIONS = [(re.compile(r'(?i)\b'+k+r'\b'), ''.join(v.split('/')[-1].strip())) for k,v in
        {
            "ain't": "am not / are not / is not / has not / have not",
            "aren't": "are not / am not",
            "can't": "cannot",
            "can't've": "cannot have",
            "'cause": "because",
            "could've": "could have",
            "couldn't": "could not",
            "couldn't've": "could not have",
            "didn't": "did not",
            "doesn't": "does not",
            "don't": "do not",
            "hadn't": "had not",
            "hadn't've": "had not have",
            "hasn't": "has not",
            "haven't": "have not",
            "he'd": "he had / he would",
            "he'd've": "he would have",
            "he'll": "he shall / he will",
            "he'll've": "he shall have / he will have",
            "he's": "he has / he is",
            "how'd": "how did",
            "how'd'y": "how do you",
            "how'll": "how will",
            "how's": "how has / how is / how does",
            "I'd": "I had / I would",
            "I'd've": "I would have",
            "I'll": "I shall / I will",
            "I'll've": "I shall have / I will have",
            "I'm": "I am",
            "I've": "I have",
            "isn't": "is not",
            "it'd": "it had / it would",
            "it'd've": "it would have",
            "it'll": "it shall / it will",
            "it'll've": "it shall have / it will have",
            "it's": "it has / it is",
            "let's": "let us",
            "ma'am": "madam",
            "mayn't": "may not",
            "might've": "might have",
            "mightn't": "might not",
            "mightn't've": "might not have",
            "must've": "must have",
            "mustn't": "must not",
            "mustn't've": "must not have",
            "needn't": "need not",
            "needn't've": "need not have",
            "o'clock": "of the clock",
            "oughtn't": "ought not",
            "oughtn't've": "ought not have",
            "shan't": "shall not",
            "sha'n't": "shall not",
            "shan't've": "shall not have",
            "she'd": "she had / she would",
            "she'd've": "she would have",
            "she'll": "she shall / she will",
            "she'll've": "she shall have / she will have",
            "she's": "she has / she is",
            "should've": "should have",
            "shouldn't": "should not",
            "shouldn't've": "should not have",
            "so've": "so have",
            "so's": "so as / so is",
            "that'd": "that would / that had",
            "that'd've": "that would have",
            "that's": "that has / that is",
            "there'd": "there had / there would",
            "there'd've": "there would have",
            "there's": "there has / there is",
            "they'd": "they had / they would",
            "they'd've": "they would have",
            "they'll": "they shall / they will",
            "they'll've": "they shall have / they will have",
            "they're": "they are",
            "they've": "they have",
            "to've": "to have",
            "wasn't": "was not",
            "we'd": "we had / we would",
            "we'd've": "we would have",
            "we'll": "we will",
            "we'll've": "we will have",
            "we're": "we are",
            "we've": "we have",
            "weren't": "were not",
            "what'll": "what shall / what will",
            "what'll've": "what shall have / what will have",
            "what're": "what are",
            "what's": "what has / what is",
            "what've": "what have",
            "when's": "when has / when is",
            "when've": "when have",
            "where'd": "where did",
            "where's": "where has / where is",
            "where've": "where have",
            "who'll": "who shall / who will",
            "who'll've": "who shall have / who will have",
            "who's": "who has / who is",
            "who've": "who have",
            "why's": "why has / why is",
            "why've": "why have",
            "will've": "will have",
            "won't": "will not",
            "won't've": "will not have",
            "would've": "would have",
            "wouldn't": "would not",
            "wouldn't've": "would not have",
            "y'all": "you all",
            "y'all'd": "you all would",
            "y'all'd've": "you all would have",
            "y'all're": "you all are",
            "y'all've": "you all have",
            "you'd": "you had / you would",
            "you'd've": "you would have",
            "you'll": "you shall / you will",
            "you'll've": "you shall have / you will have",
            "you're": "you are",
            "you've": "you have"
        }.items()
    ]

    CONTRACTIONS2 = [(re.compile(r"(?i)([^' ])"+k+r'\b'), r'\1 '+v) for k,v in
        {
            # See NLTKWordTokenizer.ENDING_QUOTES
            "'ll": "will",
            "'re": "are",
            "'ve": "have",
            "n't": "not",
            #"'s": "is", #removed this, because it breaks possessive ending
            "'m": "am",
            "'d": "would"
        }.items()
    ]

    def tokenize(self, line:str)->list[str]:
        # only split on spaces (for pos tagging there should be no further split characters since even '=' or ':' could confuse ps tagging)
        #return line.split()

        # word tokenize does what we need, respects:
        # - quotations
        # - braces/parentheses
        # - double dashes
        # - punctuations
        # - contractions (acc. Robert MacIntyre, like 'cannot'->'can not')

        # additional contractions
        for regexp, substitution in self.CONTRACTIONS:
            line=regexp.sub(substitution, line)

        # contractions nltk could manage but make downstream tasks easier
        for regexp, substitution in self.CONTRACTIONS2:
            line=regexp.sub(substitution, line)

        # Fixing leading quotations
        # nltk keeps single quotes in front of words in case it’s an apostrophe (clitics)
        # we replace this in case it is not between word characters 
        line=re.sub(re.compile(r"\B'([^' ])"), r" ' \1",line)

        return word_tokenize(line)




from sklearn_crfsuite import CRF
import pickle
import os
from poslog.words import KnownWordsDetector, RegexTokenClassMatcher, WordKind, TokenClass
from poslog.AbstractPosTagger import AbstractPosTagger
from poslog.PosLogTokenizer import PosLogTokenizer
import logging
import re
import string

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class PosLogCRF(AbstractPosTagger):
    DEFAULT_MODEL_PATH='models'
    DEFAULT_MODEL='pos_log_upos_crf_10k_model'

    def __init__(self, model_path:str=None, make_features=None):    
        self.model_path = self._get_model_path(model_path)
        self.crf = None
        self.kwdet=KnownWordsDetector()
        self.rgtcm=RegexTokenClassMatcher()
        self.tokenizer:PosLogTokenizer=None

        #TODO
        if make_features is not None:
            self.make_features=make_features

    def train(self, X_train_tokens:list[list[str]], y_train_tags:list[list[str]])->None:
        self.crf = CRF(algorithm='lbfgs', 
                       c1=0.1, 
                       c2=0.1, 
                       max_iterations=30, #More iterations takes too long#100, 
                       all_possible_transitions=True)
        feat=[self.make_features(s) for s in X_train_tokens]
        self.crf.fit(feat, y_train_tags)
        self.save_model(self.model_path)

    def train_from_tagged_sents(self, tagged_sents:list[list[tuple[str,str]]])->None:
        X_train_tokens=[[word for word,_ in tagged_sent] for tagged_sent in tagged_sents]
        y_train_tags=[[tag for _,tag in tagged_sent] for tagged_sent in tagged_sents]
        self.train(X_train_tokens, y_train_tags)

    def _get_model_path(self, model:str=None):
        """ Returns absolute path to model file. If model_path is not given, the default model is used. """
        def rel_model_path(model:str):
            model_path=os.path.join(os.path.dirname(__file__), self.DEFAULT_MODEL_PATH)
            # make sure directory exists
            os.makedirs(model_path, exist_ok=True)
            return os.path.join(model_path, model+'.pkl')
        if model is not None:
            if os.path.isabs(model):
                return model
            else:
                return rel_model_path(model)
        else:
            # choose default model
            return rel_model_path(self.DEFAULT_MODEL)

    def save_model(self, model:str=None):
        model_path=self._get_model_path(model)
        if model_path==self._get_model_path(self.DEFAULT_MODEL):
            logger.warning("Default model path is used. Won't save it. If you want to save the model to a different location, provide a model path.")
            return
        with open(model_path, 'wb') as f:
            pickle.dump(self.crf, f)
        self.model_path=model_path
        logger.info(f"Saved model to '{model_path}'")

    def load_model(self):
        with open(self.model_path, 'rb') as f:
            self.crf = pickle.load(f)

    def pos_tag(self,tokens:list[str], tagset='upos')->list[str]:
        match tagset:
            case 'upos':
                return self.predict(tokens)
            case 'ptb':
                raise NotImplementedError("PTB tagset is not implemented yet")
            case _:
                raise ValueError(f"Unknown tagset: {tagset}")
    
    def predict(self, X:list[str])->list[str]:
        if self.crf is None:
            self.load_model()
        feat=self.make_features(X)
        return list(self.crf.predict([feat])[0])

    def predict_string(self, X:str)->list[str]:
        if self.tokenizer is None:
            self.tokenizer=PosLogTokenizer()
        tokens=self.tokenizer.tokenize(X)
        return self.predict(tokens)

    def predict_string_as_tuple(self, X:str)->list[tuple[str,str]]:
        if self.tokenizer is None:
            self.tokenizer=PosLogTokenizer()
        tokens=self.tokenizer.tokenize(X)
        tags=self.predict(tokens)
        return list(zip(tokens, tags))


    def make_features(self, words:list[str])->list[dict[str,str]]:
        features_list=[]
        for i, word in enumerate(words):

            features = {}

            features['word']=word

            kind_of_known_word = self.kwdet.kind_of_known_word(word)
            features['is_stopword'] = 1 if kind_of_known_word == WordKind.STOP_WORD else 0
            features['is_wordnet'] = 1 if kind_of_known_word == WordKind.WORD_NET else 0
            features['is_wordnet'] = 1 if kind_of_known_word == WordKind.WORD_NET or kind_of_known_word == WordKind.WORDS_DICTIONARY else 0
            features['is_domain_word'] = 1 if kind_of_known_word == WordKind.DOMAIN_WORD else 0
            features['is_number'] = 1 if kind_of_known_word == WordKind.NUMBER else 0
            features['is_unknown'] = 1 if kind_of_known_word == WordKind.UNKNOWN else 0

            token_class:TokenClass = self.rgtcm.token_class(word)
            features['word_class'] = token_class.value

            features['has_upper'] = 1 if re.search(r'[A-Z]',word) else 0

            path_regex=re.compile(r'\w*:?([\.\/\\]+[\w\-:]+)+')
            features['is-path']=1 if path_regex.fullmatch(word) else 0

            # For better distinguish between 'to' as ADP and PART
            features['is_to']=1 if word.lower() == 'to' else 0

            features['contains_number'] = 1 if re.search(r'[0-9]',word) else 0
            
            features['contains_punct']=1 if re.search(r'['+string.punctuation+']',word) else 0

            # ideas from: https://www.geeksforgeeks.org/conditional-random-fields-crfs-for-pos-tagging-in-nlp/
            features['is_first'] = i == 0
            features['is_last'] = i == len(words) - 1

            features['all_caps'] = 1 if word.upper() == word else 0
            features['all_lower'] = 1 if word.lower() == word else 0

            # Next word to better distinguish between 'to' as ADP and PART
            features['next_word']= '' if i == len(words)-1 else words[i+1]
            
            features['prev_char']= '' if i == 0 else words[i-1][-1]
            features['next_char']= '' if i == len(words)-1 else words[i+1][0]

            features['prefix-1'] = word[0]
            features['prefix-2'] = word[:2]

            features['suffix-1'] = word[-1]
            features['suffix-2'] = word[-2:]
            features['suffix-3'] = word[-3:]
            
            
            word_lower = word.lower()
            features['word.lower'] = word_lower

            # Prefixes and suffixes — useful for ADJ
            features['suffix3'] = word_lower[-3:]
            features['suffix2'] = word_lower[-2:]
            features['prefix2'] = word_lower[:2]
            features['prefix3'] = word_lower[:3]

            # Punctuation — useful to catch INTJ
            features['is_punct'] = str(word in "!?.;,")

            # Position-aware features
            if i > 0:
                features['prev_word'] = words[i - 1].lower()
                features['prev_is_upper'] = str(words[i - 1].isupper())
            if i < len(words) - 1:
                # features['next_word'] = words[i + 1].lower()
                features['next_is_upper'] = str(words[i + 1].isupper())

            # Shape-based features
            features['word_shape'] = get_shape(word)

            # particles = {'not', 'off', 'up', 'down'}
            # interjections = {'oh', 'ah', 'wow', 'hey', 'oops', 'ouch', 'ok', 'bye', 'yes'}
            adjective_suffixes = ('ous', 'ful', 'ive', 'able', 'al', 'ic', 'less', 'ish')

            features['adj_suffix_match'] = str(any(word_lower.endswith(suf) for suf in adjective_suffixes))

            noun_suffixes = ('tion', 'ment', 'ness', 'ity', 'ship', 'age', 'ism', 'ence', 'ance', 'hood', 'dom')
            features['noun_suffix'] = str(any(word_lower.endswith(suf) for suf in noun_suffixes))

            # e.g., "the big ___" → likely NOUN
            determiners = {'the', 'a', 'an', 'this', 'that', 'these', 'those'}
            features['prev_is_determiner'] = str(i > 0 and words[i-1].lower() in determiners)

            features_list.append(features)
        
        return features_list        

def get_shape(word: str) -> str:
    shape = ''
    for char in word:
        if char.isupper():
            shape += 'X'
        elif char.islower():
            shape += 'x'
        elif char.isdigit():
            shape += 'd'
        else:
            #shape += '_'
            shape += char
    return shape
